/////////////////////////////////////////////////////////////////////////////
// CHookExplorer
//
//
//    Copyright (c) 1999 Ziff Davis, Inc. All rights reserved
//    First appeared in PC Magazine, US Edition, August, 1999
//                Programmed by Mark Sweeney
//


#include "StdAfx.h"
#include "HookExplorer.h"
#include "Shlobj.h"

extern HINSTANCE g_hInstance;

typedef struct {
	PTCHAR  pClassname;
	HWND   hWndFound;
} ENUM_CLASS, *LPENUM_CLASS; 



/////////////////////////////////////////////////////////////////////////////
// CHookExplorer
CHookExplorer::CHookExplorer()
{
	m_bLocked = false;
	m_bLockFlags = false;
	m_hCBTHook = 0;
	m_hKeyboardHook = 0;
	m_PrivateMessage = 0;

	m_szDesktop[0] = 0;
	m_szDestopDirectory[0] = 0;
	m_szCommonDestopDirectory[0] = 0;

	GetDesktopDirectories( );

	m_hLockedCursor = LoadCursor( g_hInstance, MAKEINTRESOURCE( IDC_POINTER ) );
	SubclassExplorer();

}
/////////////////////////////////////////////////////////////////////////////
// ~CHookExplorer
CHookExplorer::~CHookExplorer()
{
	UnSubclassWindows( );
	RemoveThreadFromProcessList( );
	
	if( m_hLockedCursor )
		DestroyCursor( m_hLockedCursor );

}


/////////////////////////////////////////////////////////////////////////////
// 
void CHookExplorer::UnSubclassWindows( )
{
	if( m_ExplorerOldWndProc && m_hwndExplorer )
		SetWindowLong( m_hwndExplorer, GWL_WNDPROC, (LONG)m_ExplorerOldWndProc );
	
	if( m_hCBTHook )
		UnhookWindowsHookEx( m_hCBTHook );
	
	if(	m_hKeyboardHook )
		UnhookWindowsHookEx( m_hKeyboardHook );
	
	m_ListView.UnSubClass( );
	m_TreeView.UnSubClass( );

	m_hKeyboardHook = 0;	
	m_hCBTHook = 0;
	m_hwndExplorer = 0;
	
}


/////////////////////////////////////////////////////////////////////////////
// 
BOOL CALLBACK EnumProc( HWND hWnd, LPARAM lParam )
{
TCHAR sz[256];
LPENUM_CLASS pEnumClass = (LPENUM_CLASS) lParam;
 	
	GetClassName( hWnd, sz, sizeof(sz) );

	if( _tcsicmp( sz, pEnumClass->pClassname ) == 0 ) 
	{
		pEnumClass->hWndFound = hWnd;
		return( FALSE );
	}

	return( TRUE );
}

////////////////////////////////////////////////////////////////////////////
//
HWND CHookExplorer::FindWindowClassThread( PTCHAR pClassName )
{
	ENUM_CLASS enumclass;

	enumclass.pClassname = pClassName;
	enumclass.hWndFound = 0;
	EnumThreadWindows( GetCurrentThreadId(), EnumProc, (LPARAM)&enumclass );
	return( enumclass.hWndFound );
}

/////////////////////////////////////////////////////////////////////////////
// 
HWND CHookExplorer::FindWindowClass( HWND hWnd, PTCHAR pClassName )
{
	ENUM_CLASS enumclass;

	enumclass.pClassname = pClassName;
	enumclass.hWndFound = 0;
	EnumChildWindows( hWnd, EnumProc, (LPARAM)&enumclass );
	return( enumclass.hWndFound );
}

/////////////////////////////////////////////////////////////////////////////
// EnumToolbarWindowProc
//  search child windows for the ToolbarWindow32 
BOOL CALLBACK EnumToolbarWindowProc( HWND hWnd, LPARAM lParam )
{
TCHAR sz[256];
 	
	GetClassName( hWnd, sz, sizeof(sz) );

	if( lstrcmp( sz, _TEXT("ToolbarWindow32") ) == 0 ) 
	{
		EnableWindow( hWnd, (BOOL)lParam );
	}
	return( TRUE );

}


/////////////////////////////////////////////////////////////////////////////
// DisableToolbarWindows 
// Disable all ToolbarWindow32 windows in explorer
void CHookExplorer::DisableToolbarWindows( )
{
	BOOL bDisable = ( ( m_bLockFlags == EXPLORER_LOCKED ) && m_bLocked );
	EnumChildWindows( m_hwndExplorer, (WNDENUMPROC)EnumToolbarWindowProc, (LPARAM)!bDisable); 
}



/////////////////////////////////////////////////////////////////////////////
// GetDesktopDirectories
// get the folders that comprise the desktop
//

void CHookExplorer::GetDesktopDirectories( )
{
LONG dwSize;
LPITEMIDLIST pidl;   
LPMALLOC pMalloc;


	if( SUCCEEDED( SHGetMalloc( &pMalloc ) ) ) 
	{
		SHGetSpecialFolderLocation( NULL, CSIDL_DESKTOPDIRECTORY, &pidl );
		SHGetPathFromIDList( pidl, m_szDestopDirectory );

		SHGetSpecialFolderLocation( NULL, CSIDL_COMMON_DESKTOPDIRECTORY, &pidl );
		SHGetPathFromIDList( pidl, m_szCommonDestopDirectory );

		SHGetSpecialFolderLocation( NULL, CSIDL_DESKTOP, &pidl );
		SHGetPathFromIDList( pidl, m_szDesktopDir );

		pMalloc->Free( pidl );
		pMalloc->Release(  );
	}
	
	
	//m_szDesktop will contain the text 'Desktop', I'm not sure if this is the correct way
	//to get the text 'Desktop'.  There doesn't seem to be a shell function that returns this
	//string so I grab it from the CLSID.   Interesting, some of the shell file functions will
	//return the string 'Desktop' if you pass a "" or "." as the filename to open.  I did not
	//have time to investigate this for all platforms, versions, locales, and so on....
	dwSize = sizeof( m_szDesktop );
	RegQueryValue( HKEY_CLASSES_ROOT, _TEXT("CLSID\\{00021400-0000-0000-C000-000000000046}"), m_szDesktop, &dwSize );	

	if( m_szCommonDestopDirectory[0] == 0 )
		 _tcscpy(m_szCommonDestopDirectory, _T("*") );

	if( m_szDestopDirectory[0] == 0 )
		_tcscpy( m_szDestopDirectory, _T("*") );

	if( m_szDesktopDir[0] == 0 )
		_tcscpy( m_szDesktopDir, _T("*") ); 
		
	if( m_szDesktop[0] == 0 )
		_tcscpy( m_szDesktop, _T("Desktop") ); 

	_tcsupr( m_szDestopDirectory );
	_tcsupr( m_szCommonDestopDirectory );
	_tcsupr( m_szDesktopDir );
	_tcsupr( m_szDesktop );

}

/////////////////////////////////////////////////////////////////////////////
// DirectoryChange
// The user has switched directories,  if the dir is part of the desktop 
// namespace then set the m_bLocked member
//
void CHookExplorer::DirectoryChange( LPCTSTR pDirectory )
{
	m_bLocked = FALSE;

	TCHAR szDir[MAX_PATH];
	
	_tcsncpy( szDir, pDirectory, MAX_PATH );
	szDir[MAX_PATH-1] = 0;	//force null terminate
	_tcsupr( szDir );

	m_bLocked  = FALSE;	
	if( _tcsstr( szDir, m_szDestopDirectory ) )
		m_bLocked = TRUE;
	else
		if( _tcsstr( szDir, m_szCommonDestopDirectory ) )
			m_bLocked = TRUE;
		 else
	 		if( _tcsstr( szDir, m_szDesktopDir ) )
				m_bLocked = TRUE;
			else
				if( _tcsicmp( szDir, m_szDesktop ) == 0)
					m_bLocked = TRUE;


	//need to wait for SHELLDLL_DefView to be created
	if( m_PrivateMessage == 0 )
	{
		m_PrivateMessage = RegisterWindowMessage( _T(SZ_APPNAME) );

		HWND hwnd = FindWindow( _T("Progman"), NULL );
		if( hwnd ) {
			HWND hShell = FindWindowClass( hwnd, _T("SHELLDLL_DefView") );
			if( hShell )
				m_bLockFlags = SendMessage( hShell, m_PrivateMessage, 0 , 0 );
			else
				m_PrivateMessage = 0;
			DEBUGOUT( _T("PrivateMessage returned %x LockFlags = %x"), hShell, m_bLockFlags );

			m_bLockFlags &= EXPLORER_LOCKED;
		}
	}

	DisableToolbarWindows( );

}


////////////////////////////////////////////////////////////////////////////
// SubclassExplorer
// subclass this instance of explorer
//
int CHookExplorer::SubclassExplorer( )
{
	
	// Get the HWND of the Explorer's window
	m_hwndExplorer = FindWindowClassThread( _T("ExploreWClass") );   

	//win95
	if( !m_hwndExplorer )
		m_hwndExplorer = FindWindowClassThread( _T("IEFrame") );   	   

	if( !m_hwndExplorer )
		m_hwndExplorer = FindWindowClassThread( _T("CabinetWClass") );   	   
  
	if( !IsWindow( m_hwndExplorer ) )
	{
		ATLTRACE( "Error finding explorer hwnd\n" );
		return FALSE;
	} 

	//if memory allocation error just get out!!
	if( AddToProcessList( this ) == 0 )
		return S_OK;
	
	//set a keyboard hook
	if( m_hKeyboardHook == 0 )
		m_hKeyboardHook = SetWindowsHookEx( WH_KEYBOARD, KeyboardProc, g_hInstance, GetCurrentThreadId());

	//Subclass the listview (left) and treeview (right), if it exists, of explorer
	//Yes, the treeview won't show in the CBT proc!!!!
	HWND hwnd = FindWindowClass( m_hwndExplorer, _TEXT("SysListView32") );
	if( hwnd ) 
		m_ListView.SubClass( hwnd );
		
	hwnd = FindWindowClass(m_hwndExplorer, _TEXT("SysTreeView32") );
	if( hwnd ) 
		m_TreeView.SubClass( hwnd );

	DisableToolbarWindows( );

	//We need a CBT hook since explorer will destroy and recreate the listview whenever
	//the user switches to another folder
	if( !m_hCBTHook )
		m_hCBTHook = SetWindowsHookEx( WH_CBT, CBTProc, g_hInstance, GetCurrentThreadId() );

	 m_ExplorerOldWndProc = reinterpret_cast<WNDPROC>(SetWindowLong(
                                 m_hwndExplorer, GWL_WNDPROC,
                                 reinterpret_cast<LONG>(ExplorerWndProc)));
	  
   return S_OK;
}




////////////////////////////////////////////////////////////////////////////
//ExplorerWndProc
//
LRESULT CALLBACK CHookExplorer::ExplorerWndProc(
                        HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{

	//get our 'this' pointer
	CHookExplorer* pThis = (CHookExplorer*)Get_pThis( );

	if( pThis == NULL )
	{
		DEBUGOUT( _T("INVALID PTHIS CHookExplorer") );
		return FALSE;
	}


	//the user has locked or unlocked the desktop
	if( uMsg == pThis->m_PrivateMessage )
	{
		pThis->m_bLockFlags = wParam;	
		pThis->DisableToolbarWindows( );
	}

	//if locked, display a cursor with a pad lock
	if( ( pThis->m_bLockFlags == EXPLORER_LOCKED ) && pThis->m_bLocked  )
	{
		switch( uMsg )
		{
			case WM_SETCURSOR:
			{
				WORD hittest = LOWORD( lParam );
				if( hittest == HTCLIENT )
				{
					SetCursor( pThis->m_hLockedCursor );
					return 1;
				}
			}
			break;
			
			case WM_COMMAND:
				return 0;
   		}
	}

   return CallWindowProc( pThis->m_ExplorerOldWndProc, hwnd, uMsg, wParam, lParam);
}



////////////////////////////////////////////////////////////////////////////
//CBTProc
//watch for new windows created by explorer, unsubclass the window if it is
//being destroyed
//
LRESULT CALLBACK CHookExplorer::CBTProc( int nCode, WPARAM wParam, LPARAM lParam )
{
	TCHAR szClass[MAX_PATH];
	CHookExplorer* pThis = Get_pThis( );

	if( pThis == NULL )
	{
		DEBUGOUT( _T("INVALID PTHIS CHookExplorer::CBTProc" ) );
		return FALSE;
	}



	if( pThis && nCode >= 0 )
	{
		switch( nCode )
		{
			case HCBT_CREATEWND:
			{
				HWND hwnd = (HWND)wParam;
				GetClassName( hwnd, szClass, sizeof( szClass )-1 );

				if( !lstrcmpi( szClass, _TEXT("ToolbarWindow32") ) )
					pThis->DisableToolbarWindows( );

				if( !lstrcmpi( szClass, _TEXT("SysListView32") ) )
					pThis->m_ListView.SubClass( hwnd );
			
				if( !lstrcmpi( szClass, _TEXT("SysTreeView32") ) )
					pThis->m_TreeView.SubClass( hwnd );

				if( (pThis->m_bLockFlags == EXPLORER_LOCKED ) && pThis->m_bLocked )
					if( !lstrcmpi( szClass, _TEXT("Edit") ) )
						return 1;	//destroy the window
			
			}
			break;

			case HCBT_DESTROYWND:
			{
				HWND hwnd = (HWND)wParam;
				
				GetClassName( hwnd, szClass, sizeof( szClass )-1 );

				if( hwnd == pThis->m_ListView.GetHwnd() )
					pThis->m_ListView.UnSubClass( );
				
				if( hwnd == pThis->m_TreeView.GetHwnd() )
					pThis->m_TreeView.UnSubClass( );

				if( hwnd ==	pThis->m_hwndExplorer )	
					pThis->UnSubclassWindows( );
			}
			break;
		}	
	
	}

   return CallNextHookEx(pThis->m_hCBTHook, nCode, wParam, lParam);
}


////////////////////////////////////////////////////////////////////////////
//KeyboardProc
//trap destructive key sequences 
LRESULT CALLBACK CHookExplorer::KeyboardProc(int nCode, WPARAM wParam, LPARAM lParam)
{

	CHookExplorer*  pThis = (CHookExplorer*)Get_pThis( );

	if( pThis == NULL )
		return FALSE;
	

	/////////////////////////////////////////////
	//THIS IS GETTING KEYSTROKES FOR ALL THREADS
	/////////////////////////////////////////////
	if( ( pThis->m_bLockFlags == EXPLORER_LOCKED ) && pThis->m_bLocked && nCode >= 0  )
	{
		switch( wParam )
		{
			case VK_DELETE:
			case VK_CONTROL:
			case VK_APPS:	//app menu
			case VK_F2:
				return 1;
			case 'V':
			case 'X':
				if( GetKeyState( VK_CONTROL ) < 0 ) //1.1
					return 1;			
		}		
   }

   return CallNextHookEx( pThis->m_hKeyboardHook, nCode, wParam, lParam );

}



